# GPFgrowth is algorithm to mine the partial periodic frequent pattern in temporal database.
#
#
# **Importing this algorithm into a python program**
#
#           from PAMI.partialPeriodicFrequentPattern.basic import GPFgrowth as alg
#
#           iFile = 'sampleTDB.txt'
#
#           minSup = 0.25 # can be specified between 0 and 1
#
#           maxPer = 300 # can  be specified between 0 and 1
#
#           minPR = 0.7 # can  be specified between 0 and 1
#
#           obj = alg.GPFgrowth(inputFile, minSup, maxPer, minPR, sep)
#
#           obj.mine()
#
#           partialPeriodicFrequentPatterns = obj.getPatterns()
#
#           print("Total number of partial periodic Patterns:", len(partialPeriodicFrequentPatterns))
#
#           obj.save(oFile)
#
#           Df = obj.getPatternInDf()
#
#           memUSS = obj.getMemoryUSS()
#
#           print("Total Memory in USS:", memUSS)
#
#           memRSS = obj.getMemoryRSS()
#
#           print("Total Memory in RSS", memRSS)
#
#           run = obj.getRuntime()
#
#           print("Total ExecutionTime in seconds:", run)
#

__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

import deprecated
from PAMI.partialPeriodicFrequentPattern.basic.abstract import *

orderOfItem = {}


class _Node(object):
    """
    A class used to represent the node of frequentPatternTree

    :**Attributes**:    - **item** (*int or None*) -- *Storing item of a node.*
                        - **timeStamps** (*list*) -- *To maintain the timestamps of a database at the end of the branch.*
                        - **parent** (*list*) -- *To maintain the parent of every node.*
                        - **children** (*list*) -- *To maintain the children of a node.*

    :**Methods**:    -**addChild(itemName)** -- *Storing the children to their respective parent nodes.*
    """

    def __init__(self, item, locations, parent=None):
        self.item = item
        self.locations = locations
        self.parent = parent
        self.children = {}

    def addChild(self, item, locations):
        """
        This method takes an item and locations as input, adds a new child node
        if the item does not already exist among the current node's children, or
        updates the locations of the existing child node if the item is already present.

        :param item: Represents the distinct item to be added as a child node.
        :type item: Any
        :param locations: Represents the locations associated with the item.
        :type locations: list
        :return: The child node associated with the item.
        :rtype: _Node
        """
        if item not in self.children:
            self.children[item] = _Node(item, locations, self)
        else:
            self.children[item].locations = locations + self.children[item].locations
            
        return self.children[item]

    def traverse(self):
        """
        This method constructs a transaction by traversing from the current node to the root node, collecting items along the way.

        :return: A tuple containing the transaction and the locations associated with the current node.
        :rtype: tuple(list, Any)
        """
        transaction = []
        locs = self.locations
        node = self.parent
        while node.parent is not None:
            transaction.append(node.item)
            node = node.parent
        return transaction[::-1], locs


class GPFgrowth(partialPeriodicPatterns):
    """
    **About this algorithm**

    :**Description**: GPFgrowth is algorithm to mine the partial periodic frequent pattern in temporal database.
    
    :**Reference**: R. Uday Kiran, J.N. Venkatesh, Masashi Toyoda, Masaru Kitsuregawa, P. Krishna Reddy, Discovering partial periodic-frequent patterns in a transactional database,
                  Journal of Systems and Software, Volume 125, 2017, Pages 170-182, ISSN 0164-1212, https://doi.org/10.1016/j.jss.2016.11.035.

    :**parameters**:    - **iFile** (*str*) -- *Name of the Input file to mine complete set of correlated patterns.*
                        - **oFile** (*str*) -- *Name of the output file to store complete set of correlated patterns.*
                        - **minSup** (*int or float or str*) -- *The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.*
                        - **minPR** (*str*) -- *Controls the maximum number of transactions in which any two items within a pattern can reappear.*
                        - **maxPer** (*str*) -- *Controls the maximum number of transactions in which any two items within a pattern can reappear.*
                        - **sep** (*str*) -- *This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.*

    :**Attributes**:    - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** (*float*) -- *To store the total amount of RSS memory consumed by the program.*
                        - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the completion time of the mining process.*
                        - **minSup** (*int*) -- *The user given minSup.*
                        - **maxPer** (*int*) -- *The user given maxPer.*
                        - **minPR** (*int*) -- *The user given minPR.*
                        - **finalPatterns** (*dict*) -- *It represents to store the pattern.*

    :**Methods**:           - **mine()** -- *Mining process will start from here.*
                        - **getPatterns()** -- *Complete set of patterns will be retrieved with this function.*
                        - **storePatternsInFile(ouputFile)** -- *Complete set of frequent patterns will be loaded in to an output file.*
                        - **getPatternsAsDataFrame()** -- *Complete set of frequent patterns will be loaded in to an output file.*
                        - **getMemoryUSS()** -- *Total amount of USS memory consumed by the mining process will be retrieved from this function.*
                        - **getMemoryRSS()** -- *Total amount of RSS memory consumed by the mining process will be retrieved from this function.*
                        - **getRuntime()** -- *Total amount of runtime taken by the mining process will be retrieved from this function.*

    **Execution methods**

    **Terminal command**

    .. code-block:: console

      Format:

      (.venv) $ python3 GPFgrowth.py <inputFile> <outputFile> <minSup> <maxPer> <minPR>

      Example Usage:

      (.venv) $ python3 GPFgrowth.py sampleTDB.txt output.txt 0.25 300 0.7

    .. note:: minSup can be specified in support count or a value between 0 and 1.


    **Calling from a python program**

    .. code-block:: python

            from PAMI.partialPeriodicFrequentPattern.basic import GPFgrowth as alg

            iFile = 'sampleTDB.txt'

            minSup = 0.25 # can be specified between 0 and 1

            maxPer = 300 # can  be specified between 0 and 1

            minPR = 0.7 # can  be specified between 0 and 1

            obj = alg.GPFgrowth(inputFile, minSup, maxPer, minPR, sep)

            obj.mine()

            partialPeriodicFrequentPatterns = obj.getPatterns()

            print("Total number of partial periodic Patterns:", len(partialPeriodicFrequentPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDf()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits**

    The complete program was written by Nakamura and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """
    _partialPeriodicPatterns__iFile = ' '
    _partialPeriodicPatterns__oFile = ' '
    _partialPeriodicPatterns__sep = str()
    _partialPeriodicPatterns__startTime = float()
    _partialPeriodicPatterns__endTime = float()
    _partialPeriodicPatterns__minSup = str()
    _partialPeriodicPatterns__maxPer = str()
    _partialPeriodicPatterns__minPR = str()
    _partialPeriodicPatterns__finalPatterns = {}
    runTime = 0
    _partialPeriodicPatterns__memoryUSS = float()
    _partialPeriodicPatterns__memoryRSS = float()
    __Database = []

    def __convert(self, value):
        """
        To convert the type of user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self._maxTS * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self._maxTS * value)
            else:
                value = int(value)
        return value

    def __creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self.__Database = []
        if isinstance(self._partialPeriodicPatterns__iFile, pd.DataFrame):
            timeStamp, data = [], []
            if self._partialPeriodicPatterns__iFile.empty:
                print("its empty..")
            i = self._partialPeriodicPatterns__iFile.columns.values.tolist()
            if 'ts' or 'TS' in i:
                timeStamp = self._partialPeriodicPatterns__iFile['timeStamps'].tolist()
            if 'Transactions' in i:
                data = self._partialPeriodicPatterns__iFile['Transactions'].tolist()
            if 'Patterns' in i:
                data = self._partialPeriodicPatterns__iFile['Patterns'].tolist()
            for i in range(len(data)):
                tr = [timeStamp[i]]
                tr.append(data[i])
                self.__Database.append(tr)

        if isinstance(self._partialPeriodicPatterns__iFile, str):
            if validators.url(self._partialPeriodicPatterns__iFile):
                data = urlopen(self._partialPeriodicPatterns__iFile)
                for line in data:
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._partialPeriodicPatterns__sep)]
                    temp = [x for x in temp if x]
                    self.__Database.append(temp)
            else:
                try:
                    with open(self._partialPeriodicPatterns__iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            temp = [i.rstrip() for i in line.split(self._partialPeriodicPatterns__sep)]
                            temp = [x for x in temp if x]
                            self.__Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def startMine(self):
        self.mine()

    def _ratioCalc(self, v):
        """
        This function take input v as input and returns the ratio.

        :param v: here v is an item.
        :type v: int or float.
        :return: int or float.
        """
        ratio = self._getPerSup(v) / (len(v) + 1)

        return ratio
    
    def _getPerSup(self, arr):
        """
        This function takes the arr as input and returns locs as output

        :param arr: an array contains the items.
        :type arr: array
        :return: locs
        """
        arr = list(arr) + [self._maxTS, 0]
        arr = list(set(arr))
        arr = np.sort(arr)
        arr = np.diff(arr)

        locs = len(np.where(arr <= self._partialPeriodicPatterns__maxPer)[0])

        return locs
    
    def _construct(self, items, data):

        """
        This method filters the items based on the minimum support (minSup) and
        maximum period (maxPer). It then constructs a tree structure from the
        filtered items and data.

        :param items: A dictionary where keys are items and values are lists of timestamps.
        :type items: dict
        :param data: The dataset used to construct the tree, where each entry is a list with
                     an index followed by items.
        :type data: list of lists
        :param minSup: The minimum support threshold.
        :type minSup: int
        :param maxPer: The maximum period threshold.
        :type maxPer: int or float
        :param maxTS: The maximum timestamp.
        :type maxTS: int or float
        :param patterns: A dictionary to store the patterns discovered during the construction.
        :type patterns: dict
        :return: A tuple containing the root node of the constructed tree and a dictionary
                 of item nodes.
        :rtype: tuple(_Node, dict)
        """


        items = {k: v for k, v in items.items() if len(v) >= self._partialPeriodicPatterns__minSup}

        for item, ts in items.items():
            ratio = self._ratioCalc(ts)
            if ratio >= self._partialPeriodicPatterns__minPR:
                self._partialPeriodicPatterns__finalPatterns[tuple([item])] = [len(ts), ratio]

        root = _Node([], None, None)
        itemNodes = {}
        for line in data:
            currNode = root
            index = int(line[0])
            line = line[1:]
            line = sorted([item for item in line if item in items], key = lambda x: len(items[x]), reverse = True)
            for item in line:
                currNode = currNode.addChild(item, [index])   # heavy
                if item in itemNodes:
                    itemNodes[item].add(currNode)
                else:
                    itemNodes[item] = set([currNode])

        return root, itemNodes

    def _recursive(self, root, itemNode):
        """
        This method recursively constructs a pattern tree from the given root node,
        filtering items based on the minimum support (minSup) and maximum period (maxPer).
        It updates the patterns dictionary with the discovered patterns.

        :param root: The current root node of the pattern tree.
        :type root: _Node
        :param itemNode: A dictionary where keys are items and values are sets of nodes associated with those items.
        :type itemNode: dict
        :param minSup: The minimum support threshold.
        :type minSup: int
        :param maxPer: The maximum period threshold.
        :type maxPer: int or float
        :param patterns: A dictionary to store the patterns discovered during the recursion.
        :type patterns: dict
        :param maxTS: The maximum timestamp.
        :type maxTS: int or float
        """

        for item in itemNode:
            newRoot = _Node(root.item + [item], None, None)

            itemLocs = {}
            transactions = {}
            for node in itemNode[item]:
                transaction, locs = node.traverse()
                if len(transaction) < 1:
                    continue
                # transactions.append((transaction, locs))
                if tuple(transaction) in transactions:
                    transactions[tuple(transaction)].extend(locs)
                else:
                    transactions[tuple(transaction)] = locs

                for item in transaction:
                    if item in itemLocs:
                        itemLocs[item] += locs
                    else:
                        itemLocs[item] = list(locs)

            # Precompute getMaxPer results for itemLocs
            # maxPerResults = {item: self._getMaxPer(itemLocs[item], maxTS) for item in itemLocs if len(itemLocs[item]) >= minSup}
            maxPerResults = {item: self._ratioCalc(itemLocs[item]) for item in itemLocs if len(itemLocs[item]) >= self._partialPeriodicPatterns__minSup}

            for item in maxPerResults:
                if maxPerResults[item] >= self._partialPeriodicPatterns__minPR:
                    self._partialPeriodicPatterns__finalPatterns[tuple(newRoot.item + [item])] = [len(itemLocs[item]), maxPerResults[item]]

            # Filter itemLocs based on minSup and maxPer
            itemLocs = {k: len(v) for k, v in itemLocs.items() if len(v) >= self._partialPeriodicPatterns__minSup}

            if not itemLocs:
                continue

            newItemNodes = {}

            for transaction, locs in transactions.items():
                transaction = sorted([item for item in transaction if item in itemLocs], key = lambda x: itemLocs[x], reverse = True)
                if len(transaction) < 1:
                    continue
                currNode = newRoot
                for item in transaction:
                    currNode = currNode.addChild(item, locs)
                    if item in newItemNodes:
                        newItemNodes[item].add(currNode)
                    else:
                        newItemNodes[item] = set([currNode])

            self._recursive(newRoot, newItemNodes)

    def mine(self):
        self._partialPeriodicPatterns__startTime = time.time()
        self._partialPeriodicPatterns__finalPatterns = {}
        self.__creatingItemSets()
        
        self._partialPeriodicPatterns__minPR = float(self._partialPeriodicPatterns__minPR)
        
        self._maxTS = 0
        items = {}
        for line in self.__Database:
            index = int(line[0])
            self._maxTS = max(self._maxTS, index)
            for item in line[1:]:
                if item not in items:
                    items[item] = []
                items[item].append(index)

        self._partialPeriodicPatterns__minSup = self.__convert(self._partialPeriodicPatterns__minSup)
        self._partialPeriodicPatterns__maxPer = self.__convert(self._partialPeriodicPatterns__maxPer)

        # def _construct(self, items, data, patterns):
        root, itemNodes = self._construct(items, self.__Database)
        self._recursive(root, itemNodes)
        
        temp = {}
        for k,v in self._partialPeriodicPatterns__finalPatterns.items():
            k = list(k)
            k = "\t".join(k)
            temp[k] = v
        self._partialPeriodicPatterns__finalPatterns = temp


        self._partialPeriodicPatterns__endTime = time.time()
        self.__runTime = self._partialPeriodicPatterns__endTime - self._partialPeriodicPatterns__startTime
        process = psutil.Process(os.getpid())
        self._partialPeriodicPatterns__memoryUSS = float()
        self._partialPeriodicPatterns__memoryRSS = float()
        self._partialPeriodicPatterns__memoryUSS = process.memory_full_info().uss
        self._partialPeriodicPatterns__memoryRSS = process.memory_info().rss


    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._partialPeriodicPatterns__memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._partialPeriodicPatterns__memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.__runTime

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self.oFile = outFile
        # writer = open(self.oFile, 'w+')
        # for x, y in self._partialPeriodicPatterns__finalPatterns.items():
        #     if len(x) == 1:
        #         writer.write(f'{x[0][0]}:{y[0]}:{y[1]}\n')
        #     else:
        #         writer.write(f'{x[0][0]}')
        #         for item in x[1:]:
        #             writer.write(f'\t{item[0]}')
        #         writer.write(f':{y[0]}:{y[1]}\n')\

        with open(self.oFile, 'w') as f:
            for x, y in self._partialPeriodicPatterns__finalPatterns.items():
                # print(list(x), y)
                f.write(x + ":" + str(y[0]) + ":" + str(y[1]) + "\n")

    def getPatternsAsDataFrame(self):
        """
        Storing final periodic-frequent patterns in a dataframe

        :return: returning periodic-frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._partialPeriodicPatterns__finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodic Ratio'])
        return dataFrame
    
    def getPatterns(self):
        """
        This function returns the final partial Periodic Patterns.

        :return: dictionary
        """
        return self._partialPeriodicPatterns__finalPatterns


    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of Partial Periodic Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())

if __name__ == '__main__':
    ap = str()
    if len(sys.argv) == 6 or len(sys.argv) == 7:
        if len(sys.argv) == 7:
            ap = GPFgrowth(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5], sys.argv[6])
        if len(sys.argv) == 6:
            ap = GPFgrowth(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5])
        ap.mine()
        print("Total number of Frequent Patterns:", len(ap.getPatterns()))
        ap.save(sys.argv[2])
        print("Total Memory in USS:", ap.getMemoryUSS())
        print("Total Memory in RSS", ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", ap.getRuntime())
    else:

        print("Error! The number of input parameters do not match the total number of parameters provided")


#  ParallelPFPGrowth is one of the fundamental distributed algorithm to discover periodic-frequent patterns in a transactional database. It is based PySpark framework.
#
# *Importing this algorithm into a python program*
# --------------------------------------------------------
#
#
#             from PAMI.periodicFrequentPattern.basic import parallelPFPGrowth as alg
#
#             obj = alg.parallelPFPGrowth(iFile, minSup, maxPer, numWorkers, sep='\t')
#
#             obj.startMine()
#
#             periodicFrequentPatterns = obj.getPatterns()
#
#             print("Total number of Periodic Frequent Patterns:", len(periodicFrequentPatterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


_copyright_ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

import sys
from collections import defaultdict
from operator import add
from pyspark.sql import SparkSession
from pyspark import SparkConf, SparkContext
import time
import psutil
import os
import pandas as pd


class Node:
    """
    A class used to represent the node of frequentPatternTree

    :Attributes:

        item : int or None
            Storing item of a node
        timeStamps : list
            To maintain the timestamps of a database at the end of the branch
        parent : node
            To maintain the parent of every node
        children : list
            To maintain the children of a node
        count : int
            To maintain the count of every node

    :Methods:

        addChild(itemName)
            Storing the children to their respective parent nodes
        toString()
            To print the node
    """

    def _init_(self, item, prefix):
        """
        Initializing the Node class

        :param item: item of a node
        :param count: count of a node
        :param children: children of a node

        """
        self.item = item
        self.count = 0
        self.children = {}
        self.prefix = prefix


class Tree:
    def _init_(self):
        """
        Initializes the Tree class with a root node, a dictionary to keep track of node links, and
        a defaultdict to store item counts.

        :param root: The root node of the FP-Tree, initialized as an empty node.
        :param nodeLink: A dictionary that links nodes of the same item for efficient traversal.
        :param itemCount: A defaultdict to keep count of item frequencies across transactions.
        """
        self.root = Node(None, [])
        self.nodeLink = {}
        self.itemCount = defaultdict(int)

    def addTransaction(self, transaction, count):
        """
        Adds a transaction to the FP-Tree, updating item counts and building the tree structure.

        :param transaction: A list of items in the transaction to be added to the tree.
        :param count: The count associated with this transaction, used to increment node counts.
        :return: None
        """
        current = self.root
        for item in transaction:
            if item not in current.children:
                current.children[item] = Node(item, transaction[0:transaction.index(item)])
                current.children[item].count += count
                self.addNodeToNodeLink(current.children[item])
            else:
                current.children[item].count += count
            self.itemCount[item] += count
            current = current.children[item]

    def addNodeToNodeLink(self, node):
        """
        Adds a node to the nodeLink dictionary, which maintains a list of nodes per item
        for easy lookup of nodes with the same item.

        :param node: The node to be added to the nodeLink structure.
        :return: None
        """
        if node.item not in self.nodeLink:
            self.nodeLink[node.item] = [node]
        else:
            self.nodeLink[node.item].append(node)

    def generateConditionalTree(self, item):
        """
        Generates a conditional FP-Tree for the specified item, which includes only paths
        containing that item and their counts.

        :param item: The item for which to generate a conditional tree.
        :return: A new Tree object representing the conditional FP-Tree.
        """
        tree = Tree()
        for node in self.nodeLink[item]:
            tree.addTransaction(node.prefix, node.count)
        return tree


class Parallel_PPFP:
    """

    A class used to represent the periodic frequent pattern tree

    :Attributes:

        root : node
            To maintain the root of the tree
        summaries : dict
            To maintain the summary of the tree

    :Methods:

        add(basket, tid, count)
            To add the basket to the tree
        getTransactions()
            To get the transactions of the tree
        merge(tree)
            To merge the tree
        project(itemId)
            To project the tree
        satisfyPer(tids, maxPer, numTrans)
            To satisfy the periodicity constraint
        extract(minCount, maxPer, numTrans, isResponsible = lambda x:True)
            To extract the periodic frequent patterns


    """

    def _init_(self, inputData, minSup, maxPeriod, numWorkers, sep='\t'):
        self._minSup = minSup
        self._maxPeriod = int(maxPeriod)
        self._numPartitions = int(numWorkers)
        self._startTime = 0.0
        self._endTime = 0.0
        self._finalPatterns = {}
        self._FPList = []
        self._inputData = inputData
        self._sep = sep
        self._memoryUSS = 0.0
        self._memoryRSS = 0.0
        self._lno = 0
        self.sc = None

    def startMine(self):
        """
        Start the mining process

        """
        self._minSup = self._convert(self._minSup)
        self._startTime = time.time()

        # Initialize SparkContext
        conf = SparkConf().setAppName("Parallel_PPFP").setMaster("local[*]")
        sc = SparkContext(conf=conf)

        if isinstance(self._inputData, str):  # Check if input is a file path
            rdd = sc.textFile(self._inputData, self._numPartitions).map(lambda x: x.split(self._sep)).persist()
        else:  # Assume input is already an RDD
            rdd = self._inputData.map(lambda x: x.split(self._sep)).persist()

        self._lno = rdd.count()
        self._minSup = self._convert(self._minSup)
        self._maxPeriod = self._convert(self._maxPeriod)

        # Get frequent items that meet the minimum support
        freqItems = rdd.flatMap(lambda trans: [(item, 1) for item in trans]) \
            .reduceByKey(add) \
            .filter(lambda x: x[1] >= self._minSup) \
            .sortBy(lambda x: x[1], ascending=False) \
            .collect()

        self._finalPatterns = dict(freqItems)
        self._FPList = [x[0] for x in freqItems]
        rank = dict([(item, index) for (index, item) in enumerate(self._FPList)])

        # Generate conditional transactions by partition
        workByPartition = rdd.flatMap(lambda x: self.genCondTransaction(x, rank)).groupByKey()

        # Fold data by key to create individual trees per partition
        trees = workByPartition.foldByKey(Tree(), lambda tree, data: self.buildTree(tree, data))

        # Generate periodic frequent patterns from the generated trees
        freqPatterns = trees.flatMap(lambda tree_tuple: self.genPeriodicFrequentPatterns(tree_tuple))

        result = freqPatterns.map(
            lambda ranks_count: (tuple([self._FPList[z] for z in ranks_count[0]]), ranks_count[1])) \
            .collect()

        # Update final patterns with results from frequent patterns
        self._finalPatterns.update(dict(result))

        # Track execution time and memory usage
        self._endTime = time.time()
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

        # Stop the SparkContext
        sc.stop()

        print("Periodic frequent patterns were generated successfully using Parallel Periodic FPGrowth algorithm")

    def _convert(self, value):
        """
        to convert the type of user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        elif type(value) is float:
            value = (self._lno * value)
        elif type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self._lno * value)
            else:
                value = int(value)
        else:
            print("minSup is not correct")
        return value

    def getPartitionId(self, value):
        """
        Get the frequent items from the database

        :param data: database
        :return: frequent items

        """
        return value % self._numPartitions

    def genCondTransaction(self, trans, rank):
        """
        Get the conditional transactions from the database

        :param tid: timestamp of a database
        :param basket: basket of a database
        :param rank: rank of a database
        :param nPartitions: number of partitions
        """
        newTrans = [rank[item] for item in trans if item in rank.keys()]
        newTrans = sorted(newTrans)
        condTrans = {}
        for i in reversed(newTrans):
            partition = self.getPartitionId(i)
            if partition not in condTrans:
                condTrans[partition] = newTrans[:newTrans.index(i) + 1]
        return [x for x in condTrans.items()]

    @staticmethod
    def buildTree(tree, data):
        """
        Constructs the FP-Tree by adding transactions from the dataset.

        :param tree: The FP-Tree being constructed.
        :param data: A list of transactions, where each transaction is a list of items.
        :return: The constructed FP-Tree with transactions added.
        """
        for trans in data:
            tree.addTransaction(trans, 1)
        return tree

    def genPeriodicFrequentPatterns(self, tree_tuple):
        """
        Generates periodic frequent patterns for a partitioned FP-Tree.

        :param tree_tuple: A tuple containing partition ID and an FP-Tree.
        :return: A generator of periodic frequent patterns and their support counts.
        """
        itemList = sorted(tree_tuple[1].itemCount.items(), key=lambda x: x[1])
        itemList = [x[0] for x in itemList]
        freqPatterns = {}
        for item in itemList:
            if self.getPartitionId(item) == tree_tuple[0]:
                freqPatterns.update(self.genPeriodicPatterns(item, [item], tree_tuple[1]))
        return freqPatterns.items()

    def genPeriodicPatterns(self, item, prefix, tree):
        """
        Recursively generates periodic patterns for an item, based on a given prefix
        and an FP-Tree structure.

        :param item: The current item being analyzed.
        :param prefix: The prefix pattern leading up to this item.
        :param tree: The FP-Tree from which to generate patterns.
        :return: A dictionary of periodic frequent patterns and their support counts.
        """
        condTree = tree.generateConditionalTree(item)
        freqPatterns = {}
        freqItems = {}
        for i in condTree.nodeLink.keys():
            freqItems[i] = 0
            for node in condTree.nodeLink[i]:
                freqItems[i] += node.count

        # Apply periodicity check
        freqItems = {key: value for key, value in freqItems.items() if value >= self._minSup}
        for i in freqItems:
            pattern = prefix + [i]
            freqPatterns[tuple(pattern)] = freqItems[i]
            freqPatterns.update(self.genPeriodicPatterns(i, pattern, condTree))
        return freqPatterns

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """
        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """
        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """
        dataFrame = pd.DataFrame(list(self._finalPatterns.items()), columns=['Patterns', 'Support'])
        return dataFrame

    def save(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        with open(self._oFile, 'w+') as writer:
            for x, y in self._finalPatterns.items():
                if type(x) == tuple:
                    pattern = " ".join(map(str, x))
                    s1 = f"{pattern}:{y}"
                else:
                    s1 = f"{x}:{y}"
                writer.write(f"{s1}\n")

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns

        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of Periodic Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    if len(sys.argv) == 6:
        inputData = sys.argv[1] if sys.argv[1].lower().endswith('.txt') else sc.textFile(sys.argv[1])
        pp_fp = Parallel_PPFP(inputData, sys.argv[2], sys.argv[3], sys.argv[4], sys.argv[5])
        pp_fp.startMine()
        finalPatterns = pp_fp.getPatterns()
        print("Total number of Periodic Frequent Patterns:", len(finalPatterns))
        pp_fp.save(sys.argv[2])
        memUSS = pp_fp.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = pp_fp.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = pp_fp.getRuntime()
        print("Total ExecutionTime in ms:", run)
    else:
        print("Error! The number of input parameters does not match the total number of parameters provided")


# The goal of the below is to obtain flat transactions from the output of subgraph mining
# Flat transactions are transactions that contain the list of subgraphs that are present in a graph
#
#   from PAMI.extras.graph import flatTransactions as ft
#
#   obj = ft.FlatTransactions()
#
#   flatTransactions = obj.getFlatTransactions(fidGidDictMap)
#   (fidGidDictMap is a list of dictionaries with keys 'FID' and 'GIDs'
#   FID is subgraph/fragment ID and GIDs are the graph IDs that contain the subgraph)
#
#   obj.saveFlatTransactions(oFile)

class Subgraphs2FlatTransactions:

    def __init__(self):
        self.flatTransactions = {}

    def getFlatTransactions(self, fidGidDictMap):
        """
        fidGidMap is a list of dictionaries with keys 'FID' and 'GIDs'
        An example of this type of output is: getSubgraphGraphMapping in GSpan class
        from subgraphMining/basic/gspan.py in PAMI
        """
        graphToSubgraphs = {}

        for mapping in fidGidDictMap:
            fid = mapping['FID']
            gids = mapping['GIDs']

            for gid in gids:
                if gid not in graphToSubgraphs:
                    graphToSubgraphs[gid] = []
                graphToSubgraphs[gid].append(fid)

        for gid in graphToSubgraphs:
            graphToSubgraphs[gid] = sorted(set(graphToSubgraphs[gid]))

        self.flatTransactions = graphToSubgraphs
        return self.flatTransactions

    def saveFlatTransactions(self, oFile):
        """
        Save the available flat transactions to a file
        """
        with open(oFile, 'w') as f:
            for _, fids in self.flatTransactions.items():
                f.write(f"{' '.join(map(str, fids))}\n")



__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import smtplib, ssl


from email.message import EmailMessage

class gmail():

    def __init__(self, userName: str, password: str) -> None:
        self.userName = userName
        self.password = password


    def send(self, toAddress: str, subject: str, body: str) -> None:
        smtp_server = smtplib.SMTP('smtp.gmail.com', 587)
        try:

            smtp_server.starttls()
            smtp_server.login(self.userName, self.password)

            message = EmailMessage()
            message.set_content(body)

            message['Subject'] = subject
            message['From'] = self.userName
            message['To'] = toAddress

            smtp_server.send_message(message)
        except Exception as e:
            # Print any error messages to stdout
            print(e)
        finally:
            smtp_server.quit()


# Parquet2CSV is a code used to converts the input Parquet file into a CSV file by the specified separator.
#
# **Importing this algorithm into a python program**
#
#             from PAMI.extras.convert import Parquet2CSV as p2c
#
#             obj = p2c.Parquet2CSV(input.parquet, sampleDB.csv, sep)
#
#             obj.convert()
#
#             obj.printStats()
#

__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import pandas as pd
import os
import psutil
import time

class Parquet2CSV:
    """
        **About this algorithm**

        :**Description**:  This class is to convert Parquet format into CSV file.

        :**Reference**:

        :**Parameters**:    - **inputFile** (*str*) -- *Path to the input Parquet file.*
                            - **outputFile** (*str*) -- *Path to the output CSV file.*
                            - **sep** (*str*) -- *This variable is used to distinguish items from one another. The default seperator is tab space. However, the users can override their default separator.*

        :**Attributes**:    - **getMemoryUSS** (*int*) -- *Returns the memory used by the process in USS.*
                            - **getMemoryRSS** (*int*) -- *Returns the memory used by the process in RSS.*
                            - **getRuntime()** (*float*) -- *Returns the time taken to execute the conversion.*
                            - **printStats()** -- * Prints statistics about memory usage and runtime.*

        :**Methods**:       - **convert()** -- *Reads the Parquet file, converts it to a CSV file, and tracks memory usage and runtime.*

        **Execution methods**

        **Terminal command**

        .. code-block:: console

          Format:

          (.venv) $ python3 _CSV2Parquet.py <inputFile> <outputFile> <sep>

          Example Usage:

          (.venv) $ python3 _CSV2Parquet.py output.parquet sampleDB.csv \t


        **Calling from a python program**

        .. code-block:: python

                import PAMI.extras.convert.Parquet2CSV as pc

                inputFile = 'output.parquet'

                sep = "\t"

                outputFile = 'sampleDB.csv'

                obj = pc.Parquet2CSV(inputFile, outputFile, sep)

                obj.convert()

                obj.printStats()


        **Credits**

        The complete program was written by P. Likhitha  and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """
    def __init__(self, inputFile, outputFile, sep):
        self.inputFile = inputFile
        self.outputFile = outputFile
        self.sep = sep

    def convert(self):
        """
        This function converts the input Parquet file into a CSV file where each row is joined by the specified separator and written to the output file.
        """
        self.start = time.time()
        df = pd.read_parquet(self.inputFile)

        with open(self.outputFile, "w") as f:
            for i in range(len(df)):
                f.write(self.sep.join(df.iloc[i]) + "\n")

        self.end = time.time()

        self.pid = os.getpid()
        process = psutil.Process(self.pid)
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """
        Returns the memory used by the process in USS (Unique Set Size).

        :return: The amount of memory (in bytes) used exclusively by the process
        :rtype: int
        """
        return self.memoryUSS

    def getMemoryRSS(self):
        """
        Returns the memory used by the process in RSS (Resident Set Size).

        :return: The total memory (in bytes) used by the process in RAM.
        :rtype: int
        """
        return self.memoryRSS

    def getRuntime(self):
        """
        Returns the time taken to complete the Parquet to CSV conversion.

        :return: The runtime of the conversion process in seconds.
        :rtype: float
        """
        return self.end - self.start

    def printStats(self):
        """
        Prints the resource usage statistics including memory consumption (USS and RSS) and the runtime.

        :return: Prints memory usage and runtime to the console.
        """
        print("Memory usage (USS):", self.memoryUSS)
        print("Memory usage (RSS):", self.memoryRSS)
        print("Runtime:", self.end - self.start)


sep = "\t"
outputFile = "output.csv"
obj = Parquet2CSV("output.parquet", outputFile, sep)
obj.convert()
obj.printStats()



# Fuzzy Frequent  Pattern-Miner is desired to find all  frequent fuzzy patterns which is on-trivial and challenging problem to its huge search space.we are using efficient pruning techniques to reduce the search space.
#
# **Importing this algorithm into a python program**
#
#             from PAMI.fuzzyFrequentPattern import FFIMiner as alg
#
#             iFile = 'sampleTDB.txt'
#
#             minSup = 0.25 # can be specified between 0 and 1
#
#             obj = alg.FFIMiner(iFile, minSup, sep)
#
#             obj.mine()
#
#             fuzzyFrequentPattern = obj.getPatterns()
#
#             print("Total number of Fuzzy Frequent Patterns:", len(fuzzyFrequentPattern))
#
#             obj.save("outputFile")
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

from PAMI.fuzzyFrequentPattern.basic import abstract as _ab
from typing import List, Dict, Tuple, Set, Union, Any, Generator
from deprecated import deprecated

class FFIMiner(_ab._fuzzyFrequentPattenrs):
    """
    **About this algorithm**

    :**Description**:   Fuzzy Frequent  Pattern-Miner is desired to find all  frequent fuzzy patterns which is on-trivial and challenging problem
                        to its huge search space.we are using efficient pruning techniques to reduce the search space.

    :**Reference**:   Lin, Chun-Wei & Li, Ting & Fournier Viger, Philippe & Hong, Tzung-Pei. (2015).
                      A fast Algorithm for mining fuzzy frequent itemsets. Journal of Intelligent & Fuzzy Systems. 29.
                      2373-2379. 10.3233/IFS-151936.
                      https://www.researchgate.net/publication/286510908_A_fast_Algorithm_for_mining_fuzzy_frequent_itemSets

    :**parameters**:    - **iFile** (*str*) -- *Name of the Input file to mine complete set of correlated patterns*
                        - **oFile** (*str*) -- *Name of the output file to store complete set of correlated patterns*
                        - **minSup** (*int or float or str*) -- *The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.*
                        - **sep** (*str*) -- *This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.*

    :**Attributes**:   - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** (*float*) -- *To store the total amount of RSS memory consumed by the program.*
                        - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the completion time of the mining process.*
                        - **itemsCnt** (*int*) -- *To record the number of fuzzy spatial itemSets generated.*
                        - **mapItemSum** (*int*) -- *To keep track of sum of Fuzzy Values of items.*
                        - **joinsCnt** (*int*) -- * To keep track of the number of ffi-list that was constructed.*
                        - **BufferSize** (*int*) -- *Represent the size of Buffer.*
                        - **itemSetBuffer** (*list*) -- *To keep track of items in buffer.*

    :**Methods**:    - **mine()** -- *Mining process will start from here.*
                     - **getPatterns()** -- *Complete set of patterns will be retrieved with this function.*
                     - **save(oFile)** -- *Complete set of frequent patterns will be loaded in to a output file.*
                     - **getPatternsAsDataFrame()** -- *Complete set of frequent patterns will be loaded in to a dataframe.*
                     - **getMemoryUSS()** -- *Total amount of USS memory consumed by the mining process will be retrieved from this function.*
                     - **getMemoryRSS()** -- *Total amount of RSS memory consumed by the mining process will be retrieved from this function.*
                     - **getRuntime()** -- *Total amount of runtime taken by the mining process will be retrieved from this function.*
                     - **convert(value)** -- *To convert the given user specified value.*
                     - **compareItems(o1, o2)** -- *A Function that sort all ffi-list in ascending order of Support.*
                     - **FSFIMining(prefix, prefixLen, FSFIM, minSup)** -- *Method generate ffi from prefix.*
                     - **construct(px, py)** -- *A function to construct Fuzzy itemSet from 2 fuzzy itemSets.*
                     - **findElementWithTID(uList, tid)** -- *To find element with same tid as given.*
                     - **WriteOut(prefix, prefixLen, item, sumIUtil)** -- *To Store the pattern.*

    **Execution methods**

    **Terminal command**

    .. code-block:: console

      Format:

      (.venv) $ python3 FFIMiner.py <inputFile> <outputFile> <minSup> <sep>

      Example Usage:

      (.venv) $ python3  FFIMiner.py sampleTDB.txt output.txt 6

    .. note:: minSup can be specified in support count or a value between 0 and 1.


    **Calling from a python program**

    .. code-block:: python

            from PAMI.fuzzyFrequentPattern import FFIMiner as alg

            iFile = 'sampleTDB.txt'

            minSup = 0.25 # can be specified between 0 and 1

            obj = alg.CoMine(iFile, minSup, sep)

            obj.mine()

            fuzzyFrequentPattern = obj.getPatterns()

            print("Total number of Fuzzy Frequent Patterns:", len(fuzzyFrequentPattern))

            obj.save("outputFile")

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)


    **Credits**

    The complete program was written by B.Sai Chitra and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """

    _startTime = float()
    _endTime = float()
    _minSup = str()
    _maxPer = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _fuzFile = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _sep = "\t"

    def __init__(self, iFile: str, minSup: float, sep: str="\t") -> None:
        super().__init__(iFile, minSup, sep)
        self._startTime = 0
        self._endTime = 0
        self._dbLen = 0
        self._minSup = minSup
        self._iFile = iFile
        self._sep = sep
        self._finalPatterns = {}
        self._memoryUSS = 0
        self._memoryRSS = 0

    def _creatingItemsets(self) -> None:
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self._transactions, self._fuzzyValues, self._Database = [], [], []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._transactions = self._iFile['Transactions'].tolist()
            if 'fuzzyValues' in i:
                self._fuzzyValues = self._iFile['fuzzyValues'].tolist()
            # print(self.Database)
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line = line.decode("utf-8")
                    line = line.split("\n")[0]
                    parts = line.split(":")
                    parts[0] = parts[0].strip()
                    parts[1] = parts[1].strip()
                    items = parts[0].split(self._sep)
                    quantities = parts[1].split(self._sep)
                    self._transactions.append([x for x in items])
                    self._fuzzyValues.append([float(x) for x in quantities])
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.split("\n")[0]
                            parts = line.split(":")
                            parts[0] = parts[0].strip()
                            parts[1] = parts[1].strip()
                            items = parts[0].split(self._sep)
                            quantities = parts[1].split(self._sep)
                            self._transactions.append([x for x in items])
                            self._fuzzyValues.append([float(x) for x in quantities])
                except IOError:
                    print("File Not Found")
                    quit()
    
    def startMine(self):
        self.mine()

    def _convert(self, value) -> Union[int, float]:
        """
        To convert the given user specified value

        :param value: user specified value
        :type value: int or float or str
        :return: converted value
        :rtype: int or float
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self._dbLen * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self._dbLen * value)
            else:
                value = int(value)
        return value
    
    def dfs(self, cands):
        """
        Perform depth-first search (DFS) to find frequent patterns in a database.

        This method recursively combines candidate patterns and calculates their support
        in the database, storing frequent patterns and their support counts.

        :param cands: List of candidate patterns represented as tuples.
        :type cands: list
        :return: None

        This method does not return anything explicitly, but it updates internal
        attributes `_finalPatterns` and `_Database` with frequent patterns and their
        support counts.
        """
        for i in range(len(cands)):
            newCands = []
            for j in range(i + 1, len(cands)):
                newCand = tuple(cands[i] + tuple([cands[j][-1]]))
                # print(items[cands[i]], items[cands[j]])
                newCandItems = {}
                keys = self._Database[cands[i]].keys() & self._Database[cands[j]].keys()
                for k in keys:
                    newCandItems[k] = min(self._Database[cands[i]][k], self._Database[cands[j]][k])
                count = sum(newCandItems.values())
                if count >= self._minSup:
                    newCands.append(newCand)
                    self._finalPatterns[newCand] = count
                    self._Database[newCand] = newCandItems
            if len(newCands) > 1:
                self.dfs(newCands)

    def mine(self):
        """
        Main() function start from here.
        """
        self._startTime = _ab._time.time()
        items = {}
        lineNo = 0
        self._creatingItemsets()

        self._dbLen = len(self._transactions)
        for transactions, fuzzyValues in zip(self._transactions, self._fuzzyValues):
            for item, fuzzyValue in zip(transactions, fuzzyValues):
                item = tuple([item])
                if item not in items:
                    items[item] = {}
                items[item][lineNo] = fuzzyValue
            lineNo += 1

        self._minSup = self._convert(self._minSup)
        self._Database = items.copy()

        supports = {k:sum(v.values()) for k,v in items.items()}
        supports = {k:v for k,v in supports.items() if v >= self._minSup}
        self._Database = {k:v for k,v in items.items() if k in supports}
        self._Database = {k:v for k,v in sorted(self._Database.items(), key=lambda x: sum(x[1].values()), reverse=True)}

        self._finalPatterns = supports.copy()

        cands = list(self._Database.keys())
        self.dfs(cands)

        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss


    def getPatternsAsDataFrame(self) -> _ab._pd.DataFrame:
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        # dataFrame = {}
        # data = []
        # for a, b in self._finalPatterns.items():
        #     data.append([a.replace('\t', ' '), b])
        #     dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        # dataFrame = _fp._pd.DataFrame(list([[" ".join(x), y] for x,y in self._finalPatterns.items()]), columns=['Patterns', 'Support'])
        dataFrame = _ab._pd.DataFrame(list([[" ".join(x), y] for x, y in self._finalPatterns.items()]), columns=['Patterns', 'Support'])
        return dataFrame

    def getPatterns(self) -> dict:
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def getMemoryUSS(self) -> float:
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self) -> float:
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """
        return self._memoryRSS

    def getRuntime(self) -> float:
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self._endTime - self._startTime

    def save(self, outFile) -> dict:
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        :return: dictionary of frequent patterns
        :rtype: dict
        """
        # self._oFile = outFile
        # writer = open(self._oFile, 'w+')
        # for x, y in self._finalPatterns.items():
        #     patternsAndSupport = x.strip() + ":" + str(y)
        #     writer.write("%s \n" % patternsAndSupport)
        with open(outFile, 'w') as f:
            for x, y in self._finalPatterns.items():
                x = "\t".join(x)
                f.write(f"{x}:{y}\n")

    def printResults(self) -> None:
        """
        This function is used to print the results
        """
        print("Total number of Fuzzy Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in seconds:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = FFIMiner(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = FFIMiner(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.startMine()
        _ap.mine()
        print("Total number of Fuzzy-Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in seconds:", _ap.getRuntime())
    else:
        _ap = FFIMiner('/Users/tarunsreepada/Downloads/Fuzzy_T10I4D100K.csv', 400, '\t')
        # _ap.startMine()
        _ap.mine()
        print("Total number of Fuzzy-Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save('output.txt')
        print(_ap.getPatternsAsDataFrame())
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in seconds:", _ap.getRuntime())
        print("Error! The number of input parameters do not match the total number of parameters provided")




# GSP is one of the fundamental algorithm to discover sequential frequent patterns in a transactional database.
# This program employs GSP property (or downward closure property) to  reduce the search space effectively.
# This algorithm employs breadth-first search technique  to find the complete set of frequent patterns in a transactional database.
#
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             import PAMI.sequentialPatternMining.basic.GSP as alg
#
#             obj = alg.GSP(iFile, minSup)
#
#             obj.startMine()
#
#             sequentialPatternMining = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(frequentPatterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternInDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#




__copyright__ = """
 Copyright (C)  2024 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""
import collections
import itertools
import pandas as pd
from deprecated import deprecated

from PAMI.sequentialPattern.basic import abstract as _ab

_ab._sys.setrecursionlimit(10000)

class GSP(_ab._sequentialPatterns):
    """
    :Description:

        * GSP is one of the fundamental algorithm to discover sequential frequent patterns in a transactional database.
        * This program employs GSP property (or downward closure property) to  reduce the search space effectively.
        * This algorithm employs breadth-first search technique  to find the complete set of frequent patterns in a transactional database.


    :Reference:   Srikant, R., Agrawal, R. (1996). Mining sequential patterns: Generalizations and performance improvements. In: Apers, P., Bouzeghoub, M., Gardarin, G. (eds) Advances in Database Technology — EDBT '96. EDBT 1996. Lecture Notes in Computer Science, vol 1057. Springer, Berlin, Heidelberg.
    :param  iFile: str :
                   Name of the Input file to mine complete set of  Sequential frequent patterns
    :param  oFile: str :
                   Name of the output file to store complete set of  Sequential frequent patterns
    :param  minSup: float or int or str :
                    minSup measure constraints the minimum number of transactions in a database where a pattern must appear
                    Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

            iFile : str
                Input file name or path of the input file
            oFile : str
                Name of the output file or the path of output file
            minSup: float or int or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                However, the users can override their default separator.
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            finalPatterns: dict
                Storing the complete set of patterns in a dictionary variable
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            Database : list
                To store the transactions of a database in list
            _xLenDatabase: dict
                To store the datas in different sequence separated by sequence, rownumber, length.
            _xLenDatabaseSame : dict
                To store the datas in same sequence separated by sequence, rownumber, length.
            _seqSep   :str
                separator to separate each itemset

    :Methods:

            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            savePatterns(oFile)
                Complete set of frequent patterns will be loaded in to an output file
            getPatternsAsDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            candidateToFrequent(candidateList)
                Generates frequent patterns from the candidate patterns
            frequentToCandidate(frequentList, length)
                Generates candidate patterns from the frequent patterns

    **Methods to execute code on terminal**
    -------------------------------------------
    .. code-block:: console


       Format:

       (.venv) $ python3 GSP.py <inputFile> <outputFile> <minSup>

       Example usage:

       (.venv) $ python3 GSP.py sampleDB.txt patterns.txt 10.0


               .. note:: minSup will be considered in times of minSup and count of database transactions

    **Importing this algorithm into a python program**
    ----------------------------------------------------
    .. code-block:: python

            import PAMI.sequentialPatternMining.basic.GSP as alg

            obj = alg.GSP(iFile, minSup)

            obj.startMine()

            sequentialPatternMining = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(frequentPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**
    ---------------

              The complete program was written by Suzuki Shota under the supervision of Professor Rage  Uday Kiran.

    """

    _minSup = float()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _sepSeq = "-1"
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _xLenDatabase={}
    _xLenDatabaseSame = {}
    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []

        if isinstance(self._iFile, _ab._pd.DataFrame):
            temp = []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                temp = self._iFile['Transactions'].tolist()
            if "tid" in i:
                temp2=self._iFile[''].tolist()
            addList=[]
            addList.append(temp[0])
            for k in range(len(temp)-1):
                if temp2[k]==temp[k+1]:
                    addList.append(temp[k+1])
                else:
                    self._Database.append(addList)
                    addList=[]
                    addList.append(temp[k+1])
            self._Database.append(addList)
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    temp.pop()
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sepSeq)]
                            temp = [x for x in temp if x ]

                            seq = []
                            for i in temp:
                                k = -2
                                if len(i)>1:
                                    seq.append(list(sorted(set(i.split()))))

                                else:
                                    seq.append(i)

                            self._Database.append(seq)

                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value):
        """
        To convert the user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value


    def make1LenDatabase(self):
        """
        To make 1 length frequent patterns by breadth-first search technique   and update Database to sequential database
        """


        idDatabase=[]
        for line in self._Database:
            x=[]
            for seq in line:
                
                x+=list(itertools.chain.from_iterable(line))
            
            idDatabase+=list(sorted(set(x)))


        newDatabase=collections.Counter(idDatabase)
        newDatabase={i:v for i,v in newDatabase.items() if v>=self._minSup}
        self._finalPatterns={str(i):v for i,v in newDatabase.items()}
        
    def list_split(self,L):
        """
        To convert the list to list separated by sepSeq

        :param L:list pattren in one list
        :return:  return_L:list pattern separated by each sequence  
        """
        return_L,tmp = [],[]
        for val in L:
            if val != self._sepSeq:
                tmp.append(val)
            elif val == self._sepSeq:
                return_L.append(tmp)
                tmp = []
    
        return_L.append(tmp)
        return return_L
    
    def checkPattern(self,pattern,tra):
        """
        To check the pattern is included in tra 
        param pattern:the candidate pattern 
              tra:list     one transaction in data
        :return:  0 :the pattern is not included
                  1 :the pattern is included
        """
        seqNum=0
        
        for i in tra:
            if set(pattern[seqNum]).issubset(set(i)):
                seqNum+=1
                if seqNum>=len(pattern):
                    return 1
        return 0
            
            
        
    def getSup(self,pattern):
        """
        count up the support of the pattern
        :param pattren:list the candidate pattern 
        :return:  sup:int  the support of the pattern
        """
        sup=0
        pattern=self.list_split(pattern)
        for seq in self._Database:
            
            sup+=self.checkPattern(pattern,seq)
        return sup
        
    def make2LenDatabase(self):
        """
        To make 2 length frequent patterns by joining two one length patterns by breadth-first search technique  and update xlen Database to sequential database
        :return:  nextPatterns:List  the patterns found
       
        """
        self._xLenDatabase = {}
        keyList=[i for i in self._finalPatterns.keys()]
        keyNumber=0
        nextPatterns=[]
        
        for key1 in keyList:
            newPattern1=[key1,self._sepSeq,key1]
            sup=self.getSup(newPattern1)
            if sup>=self._minSup:
                self._finalPatterns[tuple(newPattern1)]=sup
                nextPatterns.append(newPattern1)
            keyNumber+=1
            for key2 in keyList[keyNumber:]:
                newPattern1=[key1,self._sepSeq,key2]
                sup=self.getSup(newPattern1)
                if sup>=self._minSup:
                    self._finalPatterns[tuple(newPattern1)]=sup
                    nextPatterns.append(newPattern1)
                newPattern2=[key2,self._sepSeq,key1]
                sup=self.getSup(newPattern2)
                if sup>=self._minSup:
                    self._finalPatterns[tuple(newPattern2)]=sup
                    nextPatterns.append(newPattern2)
                newPattern3=list(sorted(set([key1,key2])))
                sup=self.getSup(newPattern3)
                if sup>=self._minSup:
                    self._finalPatterns[tuple(newPattern3)]=sup
                    nextPatterns.append(newPattern3)
        return nextPatterns

    def makeCandidateDatabase(self,patterns):
        """
        make the database to find new candidate
        :param pattrens:list the patterns fond before
        :return:  bothBefore:dict   the patterns have same item without last one item
                  bothAfter:dict   the patterns have same item without first one item
        """
        before={}
        after={}
        
        for pattern in patterns:
            if pattern[-2]==self._sepSeq:
                if tuple(pattern[:-2]) not in after.keys():
                    after[tuple(pattern[:-2])]=[pattern[-2:]]
                else:
                    after[tuple(pattern[:-2])].append(pattern[-2:])
            else:
                if tuple(pattern[:-1]) not in after.keys():
                    after[tuple(pattern[:-1])]=[[pattern[-1]]]
                else:
                    after[tuple(pattern[:-1])].append([pattern[-1]])
            if pattern[1]==self._sepSeq:
                if tuple(pattern[2:]) not in before.keys():
                    before[tuple(pattern[2:])]=[pattern[:2]]
                else:
                    before[tuple(pattern[2:])].append(pattern[:2])
            else:
                if tuple(pattern[1:]) not in before.keys():
                    before[tuple(pattern[1:])]=[[pattern[0]]]
                else:
                    before[tuple(pattern[1:])].append([pattern[0]])
        bothBefore={i:v for i,v in before.items() if i in after.keys()}
        bothAfter={i:v for i,v in after.items() if i in before.keys()}
        return bothBefore, bothAfter
        
    def makeCandidate(self,patterns):
        """
        make the candidate patterns
        :param pattrens:list the patterns found before
        :return:  newPatterns:list  the candidate pattern
        """
        before,after=self.makeCandidateDatabase(patterns)
        newPatterns=[]
        for common in before.keys():
            for b in before[common]:
                for a in after[common]:
                    newPattern=b+list(common)+a
                    newPatterns.append(newPattern)
        return newPatterns
                
        
            
    
    def makexLenDatabase(self,patterns):
        """
        To make 3 or more length frequent patterns from pattern which the latest word is in different seq  by depth-first search technique  and update xlenDatabase to sequential database

        :param rowLen: row length of previous patterns.
        :param bs : previous patterns without the latest one
        :param latestWord : latest word of previous patterns
        
        """
        patterns=self.makeCandidate(patterns)
        nextPatterns=[]
        for pattern in patterns:
                sup=self.getSup(pattern)
                if sup>=self._minSup:
                    self._finalPatterns[tuple(pattern)]=sup
                    nextPatterns.append(pattern)
        return nextPatterns
                    
        
       

    def startMine(self):
        """
        Frequent pattern mining process will start from here
        """
        self._Database = []
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        self.make1LenDatabase()
        nextPatterns=self.make2LenDatabase()
        while(len(nextPatterns)>0):
            nextPatterns= self.makexLenDatabase(nextPatterns)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Sequential Frequent patterns were generated successfully using GSP algorithm ")

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = str(x) + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        """
        This function is used to prnt the results
        """
        print("Total number of Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = GSP(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = GSP(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Frequent Patterns:", len(_Patterns))
        _ap.savePatterns(_ab._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")




import pandas as pd
import numpy as np
import sys
# DB2DF in this code transactional databases is converting dense dataframe.
#**Importing this algorithm in terminal**
#             python3 DB2DF inputFileName outputFileName separator
# 
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.DB2DF import DB2DF as db
#
#             obj = db.DB2DF(fileName)
#
#             obj.createDB()
#            
#             df=obj.getDF()
#
#             obj.save(outputFileName)
#
"""
Copyright (C)  2024 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
"""     
class TransactionDB2denseDF():
    """
    :Description:  DB2DF in this code transactional databases is converting dense dataframe.

    :Attributes:

        :param _iFile: str :
             It is input file name.
        :param _sep: str :
             It is separator for each values. Default is '\t'.
        :param _oFile: str :
             It is output file name. Default is 'TransactialDB'+inputFile+".csv".
        :param _Database: str :
             It is database made from file.
        :param df: DataFrame :
             It is dense DataFrame.


    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

             from PAMI.extras.DB2DF import DF2DF as db
            
             obj = db.DF2DF(fileName)

             obj.createDB()
            
             df=obj.getDF()#get the dataframe

             obj.save(outputFileName)#get the dataframe by csv
    """

    def __init__(self, inputFile,sep=" ") -> None:
        self._iFile = inputFile
        self._sep=sep
        self._Database=[]
        self.oFile = 'TransactialDB'+inputFile+".csv"
        self._df = pd.DataFrame()
    def _get_unique_list(self):
        """
        create transactional database and return outputFileName
        :type seen: list
        :return: seen
        """
        seen = []
        for items in self._Database:
            for item in items:
                if item not in seen:
                    seen.append(item)
                
        return seen
        
    
    def _creatingItemSets(self) -> None:
            """
            Storing the complete transactions of the database/input file in a database variable
            """
            self._Database = []
            if isinstance(self._iFile, pd.DataFrame):
                temp = []
                if self._iFile.empty:
                    print("its empty..")
                i = self._iFile.columns.values.tolist()
                if 'Transactions' in i:
                    self._Database = self._iFile['Transactions'].tolist()
                    self._Database = [x.split(self._sep) for x in self._Database]
                else:
                    print("The column name should be Transactions and each line should be separated by tab space or a seperator specified by the user")
            if isinstance(self._iFile, str):
                
                    try:
                        with open(self._iFile, 'r', encoding='utf-8') as f:
                            for line in f:
                                line.strip()
                                temp = [i.rstrip() for i in line.split(self._sep)]
                                temp = [x for x in temp if x]
                                self._Database.append(set(temp))
                    except IOError:
                        print("File Not Found")
                        quit()

    def makeDic(self):
        """
        create dence dataframe 
        """
        column=self._get_unique_list()
        self._df = pd.DataFrame(index=range(len(self._Database)),columns=column)
        
        self._df.fillna(0, inplace=True)
        index=0
        for items in self._Database:
            for item in items:
                self._df.loc[index,str(item)]=1
            index+=1
                
    def save(self,df,oFile):
        """
        save database in csv file
        :input 
            oFile:str    output file name
        """
        df.to_csv(oFile,index=False)
    def createDB(self):
        
        self._creatingItemSets()
        self.makeDic()
    def getDF()
        """
        get database 
        :return: df
        """
        return self._df
        
if __name__ == '__main__':

    if len(sys.argv)=3:
        obj = TransactionDB2denseDF( sys.argv[0], sys.argv[2])
        obj.createDB()
        self.save(sys.argv[1])
    elif len(sys.argv)=2:
        obj = TransactionDB2denseDF( sys.argv[0])
        obj.createDB()
        self.save(sys.argv[1])
    elif len(sys.argv)=1:
        obj = TransactionDB2denseDF( sys.argv[0])
        obj.createDB()
        self.save(obj.oFile)
    else:
        print("Err:wrong input")


# dataFrameInToFigures is used to convert the given dataframe into figures.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#     from PAMI.extras.graph import DF2Fig as fig
#
#     obj = fig.DF2Fig(idf)
#
#     obj.plotGraphsFromDataFrame("minSup", "patterns")
#
#     obj.plotGraphsFromDataFrame("minSup", "memory")
#
#     obj.plotGraphsFromDataFrame("minSup", "runtime")
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sys

import plotly.express as _px
import pandas as _pd

class DF2Fig():
    """

    :Description:   DataFrameInToFigures is used to convert the given dataframe into figures.

    :param  dataFrame:
            Name of the input dataframe

    :param algorithm:
           Specify the column name containing the algorithms

    :param xcolumn:
           Specify the name of the X-axis

    :param ycolumn:
           Specify the name of the Y-axis 
           
    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.graph import dataframeInToFigures as fig

            obj = fig.dataframeInToFigures(idf)

            obj.plotGraphsFromDataFrame("minSup", "patterns", "algorithms")

            obj.plotGraphsFromDataFrame("minSup", "memory")

            obj.plotGraphsFromDataFrame("minSup", "runtime")

    """

    def __init__(self, dataFrame: _pd.DataFrame) -> None:
        self._dataFrame = dataFrame

    def plot(self, xColumn, yColumn, algorithm=None) -> None:
        """
        To plot graphs from given dataframe

        :param xColumn: Name of the X-axis of the dataframe

        :type xColumn: str

        :param yColumn: Name of the Y-axis of the dataframe

        :type yColumn: str

        :param algorithm: Specify the column name containing the algorithms

        :type algorithm: str

        :return: None
        """
        if algorithm is None:
            fig = _px.line(self._dataFrame, x=self._dataFrame[xColumn] , y=self._dataFrame[yColumn], color=self._dataFrame.iloc[:, 0], labels={'x': xColumn, 'y': yColumn})
        else:
            fig = _px.line(self._dataFrame, x=self._dataFrame[xColumn], y=self._dataFrame[yColumn],
                           color=self._dataFrame[algorithm], labels={'x': xColumn, 'y': yColumn})

        fig.show()


# if __name__ == '__main__':
#     ab = DF2Fig(result)
#     # user can change x and y columns
#     ab.plot("minSup", "patterns")





import random as _rd


class syntheticUtilityDatabase:
    """
    This class creates a synthetic utility database.

    Attributes:
        totalTransactions (int): Number of transactions.
        numOfItems (int): Number of items.
        maxUtilRange (int): Maximum utility range.
        avgTransactionLength (int): The length of average transaction.

    Methods:
        __init__(totalTransactions, numOfItems, maxUtilRange, avgTransactionLength)
            Constructor to initialize the database parameters.
        createSyntheticUtilityDatabase(outputFile)
            Create utility database and store it in the specified output file.
        createRandomNumbers(n, targetSum)
            Generate a list of random numbers with a specified target sum.
        save(outputFile)
            Save the generated utility database to a CSV file.

    Credits:
        The complete program was written by A.Hemanth sree sai under the supervision of Professor Rage Uday Kiran.
    """

    def __init__(self, totalTransactions: int, numOfItems: int, maxUtilRange: int, avgTransactionLength: int) -> None:
        """
        Constructor to initialize the database parameters.

        Parameters:
            totalTransactions (int): Number of transactions.
            numOfItems (int): Number of items.
            maxUtilRange (int): Maximum utility range.
            avgTransactionLength (int): The length of average transaction.
        """
        self.totalTransactions = totalTransactions
        self.numOfItems = numOfItems
        self.maxUtilRange = maxUtilRange
        self.avgTransactionLength = avgTransactionLength
        self.transactions = []

    def createSyntheticUtilityDatabase(self, outputFile: str) -> None:
        """
        Create utility database and store it in the specified output file.

        Parameters:
            outputFile (str): File name or path to store the database.
        """
        if self.avgTransactionLength > self.numOfItems:
            print("Error: avgTransactionLength cannot exceed numOfItems.")
            return

        with open(outputFile, 'w') as writer:
            for _ in range(self.totalTransactions):
                length = _rd.randint(1, self.avgTransactionLength + 20)
                items = [_rd.randint(1, self.numOfItems) for _ in range(length)]
                utilities = [_rd.randint(1, self.maxUtilRange) for _ in range(length)]

                # Generating 13 random numbers with a target sum of 2000
                randomNumbers = self.createRandomNumbers(13, 2000)

                # Checking if avgTransactionLength exceeds numOfItems
                if self.avgTransactionLength > self.numOfItems:
                    print("Error: avgTransactionLength cannot exceed numOfItems.")
                    return

                st = '\t'.join(map(str, items)) + '\t:' + str(sum(utilities)) + ':'
                st1 = '\t'.join(map(str, randomNumbers)) + '\t'

                writer.write(f"{st}{st1}\n")

    def createRandomNumbers(self, n: int, targetSum: int) -> list[float]:
        """
        Generate a list of random numbers with a specified target sum.

        Parameters:
            n (int): Number of random numbers to generate.
            targetSum (int): Target sum for the generated random numbers.

        Returns:
            list: List of generated random numbers normalized and multiplied by the target sum.
        """
        randNumbers = [_rd.uniform(0, 1) for _ in range(n)]
        randSum = sum(randNumbers)
        normalizedNumbers = [num / randSum for num in randNumbers]
        result = [round(num * targetSum) for num in normalizedNumbers]
        return result

    def save(self, outputFile: str) -> None:
        """
        Save the generated utility database to a CSV file.

        Parameters:
            outputFile (str): File name or path to store the CSV file.
        """
        with open(outputFile, 'w') as f:
            for transaction in self.transactions:
                f.write('\t'.join(map(str, transaction)) + '\n')


if __name__ == "__main__":
    ap = syntheticUtilityDatabase(100000, 870, 100, 10)
    ap.createSyntheticUtilityDatabase("T10_util-12.csv")
else:
    print("Error! The number of input parameters does not match the total number of parameters provided")



# TransactionalDatabase is a collection of transactions. It only considers the data in  transactions and ignores the metadata.
#
#  **Importing this algorithm into a python program**
#  --------------------------------------------------------
#     from PAMI.extras.syntheticDataGenerator import TransactionalDatabase as db
#
#     obj = db(10, 5, 10)
#
#     obj.create()
#
#     obj.save('db.txt')
#
#     print(obj.getTransactions())
#

import numpy as np
import pandas as pd
import sys


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

class TransactionalDatabase:
    """
    :Description: TransactionalDatabase is a collection of transactions. It only considers the data in  transactions and ignores the metadata.
    :Attributes:

        numLines: int
            Number of lines
        avgItemsPerLine: int
            Average number of items per line
        numItems: int
            Total number of items

    :Methods:

        create: 
            Generate the transactional database
        save: 
            Save the transactional database to a user-specified file
        getTransactions: 
            Get the transactional database


    **Methods to execute code on terminal**
    ---------------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 TransactionalDatabase.py <numLines> <avgItemsPerLine> <numItems>

      Example Usage:

      (.venv) $ python3 TransactionalDatabase.py 50.0 10.0 100



    **Importing this algorithm into a python program**
    --------------------------------------------------------
        from PAMI.extras.syntheticDataGenerator import TransactionalDatabase as db

        obj = db.TransactionalDatabase(10, 5, 10)

        obj.create()

        obj.save('db.txt')

        print(obj.getTransactions())

    
    """

    def __init__(self, databaseSize, avgItemsPerTransaction, numItems,sep = "\t") -> None:
        """
        Initialize the transactional database with the given parameters

        :param databaseSize: total number of transactions in the database
        :type databaseSize: int
        :param avgItemsPerTransaction: average number of items per transaction
        :type avgItemsPerTransaction: int
        :param numItems: total number of items
        :type numItems: int
        :param sep: separator to distinguish the items in a transaction
        :type sep: str
        """

        self.databaseSize = databaseSize
        self.avgItemsPerTransaction = avgItemsPerTransaction
        self.numItems = numItems
        self.sep = sep
        self.db = []

    def _generateArray(self, nums, avg, maxItems) -> list:
        """
        Generate a random array of length n whose values average to m

        :param nums: number of values
        :type nums: int
        :param avg: average value
        :type avg: int
        :param maxItems: maximum value
        :type maxItems: int

        Returns:
        values: list - random array
        """

        # generate n random values
        values = np.random.randint(1, avg * 2, nums)
        sums = np.sum(values)
        weights = values / sums

        # Calculate sumRes
        sumRes = nums * avg

        # Adjust values based on weights and sumRes
        new_values = np.round(sumRes * weights).astype(int)

        # if all transactions have 0 items, add 1 item to each transaction
        for loc in np.where(new_values < 1)[0]:
            new_values[loc] += 1

        difference = sumRes - np.sum(new_values)
        if difference > 0:
            for i in range(difference):
                index = np.random.randint(0, len(new_values))
                new_values[index] += 1
        else:
            for i in range(abs(difference)):
                index = np.random.randint(0, len(new_values))
                new_values[index] -= 1

        return values

    def create(self) -> None:
        """
        Generate the transactional database with the given input parameters.
        Returns: None
        """

        values = self._generateArray(self.databaseSize, self.avgItemsPerTransaction, self.numItems)

        self.db = []
        for i in range(self.databaseSize):
            self.db.append(np.random.choice(range(1, self.numItems + 1), values[i], replace=False))

    def save(self, filename) -> None:
        """
        Save the transactional database to a file

        :param filename: name of the file
        :type filename: str
        """

        with open(filename, 'w') as f:
            for line in self.db:
                f.write(str(self.sep).join(map(str, line)) + '\n')

    def getTransactions(self, sep = "\t") -> pd.DataFrame:
        """
        Get the transactional database in dataFrame format

        Returns:
        db: pd.dataFrame - transactional database
        """
        column = "Transactions"
        db = pd.DataFrame(columns=[column])
        db[column] = [sep.join(map(str, line)) for line in self.db]
        return db
        

if __name__ == "__main__":

    if len(sys.argv) == 5:
        obj = TransactionalDatabase(int(sys.argv[1]), int(sys.argv[2]), int(sys.argv[3]))
        obj.create()
        obj.save(sys.argv[4])
    if len(sys.argv) == 6:
        obj = TransactionalDatabase(int(sys.argv[1]), int(sys.argv[2]), int(sys.argv[3]), sys.argv[4])
        obj.create()
        obj.save(sys.argv[5])
    else:
        raise ValueError("Invalid number of arguments. Args: <numLines> <avgItemsPerLine> <numItems> <filename> or Args: <numLines> <avgItemsPerLine> <numItems> <sep> <filename>")
    



#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import PAMI.sequentialSpatialPattern.basic.abstract as _ab
import sys
import copy
sys.setrecursionlimit(10000)

class PrefixSpan(_ab._sequentialSpatialPatterns):
    """
        Prifix Span is one of the fundamental algorithm to discover sequential frequent patterns in a transactional database.
        This program employs Prifix Span property (or downward closure property) to  reduce the search space effectively.
        This algorithm employs depth-first search technique to find the complete set of frequent patterns in a
        transactional database.
        Reference:
        ----------
           J. Pei, J. Han, B. Mortazavi-Asl, J. Wang, H. Pinto, Q. Chen, U. Dayal, M. Hsu: Mining Sequential Patterns by Pattern-Growth: The PrefixSpan Approach. IEEE Trans. Knowl. Data Eng. 16(11): 1424-1440 (2004)
        ----------
            iFile : str
                Input file name or path of the input file
            oFile : str
                Name of the output file or the path of output file
            minSup: float or int or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                However, the users can override their default separator.
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            finalPatterns: dict
                Storing the complete set of patterns in a dictionary variable
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            Database : list
                To store the transactions of a database in list
            maxLength:int
                to store the maximum length of sequence pattern
            maxGap   :int
                to store the maximum gap of sequence pattern
                gap means the length of interval between two itemsets
            NeighboursMap:dict
                to store the neighbor map(which place is neighbor)
        Methods:
        -------
            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            savePatterns(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            getPatternsAsDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            candidateToFrequent(candidateList)
                Generates frequent patterns from the candidate patterns
            frequentToCandidate(frequentList, length)
                Generates candidate patterns from the frequent patterns
            mapNeighbours()
                read the neighbor file and make neighbor map.

        Executing the code on terminal:
        -------------------------------
            Format:
            ------
             basic:
                python3 PrefixSpan.py <inputFile> <outputFile> <neighborFile> <minSup>
             length:
                python3 PrefixSpan.py <inputFile> <outputFile> <neighborFile> <minSup> "<sep(default="\t")>" <maxLength>
             gap:
                python3 PrefixSpan.py <inputFile> <outputFile> <neighborFile> <minSup> <sep(default="\t")> <maxLength(default=float("inf"))> <maxGap>
            Examples:
            ---------
            basic
                python3 PrefixSpan.py sampleDB.txt output.txt neighbor.txt 10.0   (minSup will be considered in times of minSup and count of database transactions)
                python3 PrefixSpan.py sampleDB.txt output.txt neighbor.txt 10     (minSup will be considered in support count or frequency)
            length
                python3 PrefixSpan.py sampleDB.txt output.txt neighbor.txt 10 "\t" 2 (find the pattern that have two or less itemsets and ten or more support)
            gap
                python3 PrefixSpan.py sampleDB.txt output.txt neighbor.txt 10 "\t" float("inf") 3
            length and gap
                python3 PrefixSpan.py sampleDB.txt output.txt neighbor.txt 10 "\t" 2 3
        Sample run of the importing code:
        ---------------------------------
            import PAMI.frequentPattern.basic.PrefixSpan as alg
            obj = alg.PrefixSpan(iFile, nFile,minSup)
            obj.startMine()
            frequentPatterns = obj.getPatterns()
            print("Total number of Frequent Patterns:", len(frequentPatterns))
            obj.savePatterns(oFile)
            Df = obj.getPatternInDataFrame()
            memUSS = obj.getMemoryUSS()
            print("Total Memory in USS:", memUSS)
            memRSS = obj.getMemoryRSS()
            print("Total Memory in RSS", memRSS)
            run = obj.getRuntime()
            print("Total ExecutionTime in seconds:", run)
        Credits:
        --------
            The complete program was written by Suzuki Shota under the supervision of Professor Rage Uday Kiran.
    """
    def __init__(self,iFile,nFile, minSup, sep="\t",maxlen=float("inf"),maxGap=float("inf")):
        super().__init__( iFile,nFile, minSup, sep)


        self._startTime = float()
        self._endTime = float()
        self._finalPatterns = {}
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._Database = []
        self._sepDatabase={}
        self._maxLength=maxlen
        self._maxGap=maxGap
        self._NeighboursMap = {}
    def _mapNeighbours(self):
        """
            A function to map items to their Neighbours
        """
        self._NeighboursMap = {}
        if isinstance(self._nFile, _ab._pd.DataFrame):
            data, items = [], []
            if self._nFile.empty:
                print("its empty..")
            i = self._nFile.columns.values.tolist()
            if 'item' in i:
                items = self._nFile['items'].tolist()
            if 'Neighbours' in i:
                data = self._nFile['Neighbours'].tolist()
            for k in range(len(items)):
                self._NeighboursMap[items[k]] = data[k]
            # print(self.Database)
        if isinstance(self._nFile, str):
            if _ab._validators.url(self._nFile):
                data = _ab._urlopen(self._nFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    if temp[0] in self._Database.keys():
                        self._NeighboursMap[temp[0]] = temp[1:]
            else:
                try:
                    with open(self._nFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._NeighboursMap[temp[0]] = temp[1:]
                except IOError:
                    print("File Not Found")
                    quit()
    def _creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []

        if isinstance(self._iFile, _ab._pd.DataFrame):
            temp = []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                temp = self._iFile['Transactions'].tolist()

            for k in temp:
                self._Database.append(set(k))
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(set(temp))
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split('-1')]
                            temp = [x for x in temp if x ]
                            temp.pop()

                            seq = []
                            for i in temp:
                                if len(i)>1:
                                   for i in list(sorted(set(i.split()))):
                                       seq.append(i)
                                   seq.append(-1)

                                else:
                                    seq.append(i)
                                    seq.append(-1)
                            self._Database.append(seq)


                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value):
        """
        To convert the user specified minSup value
        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value
    def makeNext(self,sepDatabase,startrow):
        """
         To get next pattern by adding head word to next sequence of startrow
        :param sepDatabase: dict
            what words and rows startrow have to add it
        :param startrow:
            the patterns get before
        """
        for head in sepDatabase.keys():
            if head not in self._NeighboursMap.keys():
                continue
            newrow=[i for i in startrow]
            checkrow={i for i in startrow if i!=-1}

            if len(sepDatabase[head])>=self._minSup  and checkrow.issubset(self._NeighboursMap[head]):
                if newrow!=[]:
                    newrow.append(-1)
                newrow.append(head)
                newrow.append(-1)
                if str(newrow) not in self._finalPatterns:
                    self._finalPatterns[str(newrow)]=len(sepDatabase[head])
                    give = []
                    give.append(head)
                    sepDatabase[head] = self.makeSupDatabase(sepDatabase[head], give)
                    newrow.pop()
                    self.makeSeqDatabaseSame(sepDatabase[head], newrow)
                elif len(sepDatabase[head]) > self._finalPatterns[str(newrow)]:
                    self._finalPatterns[str(newrow)] = len(sepDatabase[head])
                    give = []
                    give.append(head)
                    sepDatabase[head] = self.makeSupDatabase(sepDatabase[head], give)
                    newrow.pop()
                    self.makeSeqDatabaseSame(sepDatabase[head], newrow)



    def makeSupDatabase(self,database,head):
        """
         To delete not frequent words without words in latest sequence
        :param database: list
            database of lines have same startrow and head word
        :param head:list
         words in latest sequence
        :return: changed database

        """

        supDatabase={}
        alreadyInData=[]
        newDatabase = []

        for line in database:
            alreadyInLine = []
            for data in line:
                    if data not in alreadyInLine:
                        if data not in alreadyInData:
                            supDatabase[data]=1
                            alreadyInData.append(data)
                        else:
                            supDatabase[data]+=1
                        alreadyInLine.append(data)
        for line in database:
                newLine=[]
                for i in line:
                    if supDatabase[i]>=self._minSup or i in head:
                        if len(newLine)>1:
                            if (newLine[-1]!=-1 or i!=-1):
                                newLine.append(i)
                        else:
                            newLine.append(i)
                newDatabase.append(newLine)

        return newDatabase

    def makeNextSame(self,sepDatabase,startrow):
        """
         To get next pattern by adding head word to latest sequence of startrow
        :param sepDatabase: dict
            what words and rows startrow have to add it
        :param startrow:
            the patterns get before
        """
        for head in sepDatabase.keys():
            if head not in self._NeighboursMap.keys():
                continue
            checkrow = {i for i in startrow if i != -1}
            if len(sepDatabase[head])>=self._minSup and checkrow.issubset(self._NeighboursMap[head]):
                newrow = startrow.copy()
                newrow.append(head)
                newrow.append(-1)

                if str(newrow) not in self._finalPatterns.keys():
                    self._finalPatterns[str(newrow)]=len(sepDatabase[head])
                    if -1 in startrow:
                        give = self.getSameSeq(startrow)
                    else:
                        give = startrow.copy()
                    give.append(head)
                    sepDatabase[head] = self.makeSupDatabase(sepDatabase[head], give)
                    newrow.pop()
                    self.makeSeqDatabaseSame(sepDatabase[head], newrow)
                elif len(sepDatabase[head])>self._finalPatterns[str(newrow)]:
                    self._finalPatterns[str(newrow)] = len(sepDatabase[head])
                    if -1 in startrow:
                        give = self.getSameSeq(startrow)
                    else:
                        give = startrow.copy()
                    give.append(head)
                    sepDatabase[head] = self.makeSupDatabase(sepDatabase[head], give)
                    newrow.pop()
                    self.makeSeqDatabaseSame(sepDatabase[head], newrow)








    def makeSeqDatabaseFirst(self,database):
        """
        To make 1 length sequence dataset list which start from same word. It was stored only 1 from 1 line.
        :param database:
                To store the transactions of a database in list
        """
        startrow=[]
        seqDatabase={}

        for line in database:
            alreadyInLine=[]
            for data in range(len(line)):
                if line[data] not in alreadyInLine and line[data]!=-1:
                    if line[data] not in seqDatabase.keys():
                        seqDatabase[line[data]]=[]
                        seqDatabase[line[data]].append(line[data+1:])
                        alreadyInLine.append(line[data])
                    else:
                        seqDatabase[line[data]].append(line[data+1:])
                        alreadyInLine.append(line[data])

        if len(seqDatabase)>0:
            self.makeNext(seqDatabase,startrow)

    def serchSame(self,database,startrow,give):
        """
         To get 2 or more length patterns in same sequence.
        :param database: list
            To store the transactions of a database in list which have same startrow and head word
        :param startrow: list
            the patterns get before
        :param give: list
            the word in latest sequence of startrow
        :return:
        """
        sepDatabaseSame={}
        sepDatabaseSame[startrow[-1]]=[]
        for line in database:
            addLine=0
            i=0
            if len(line)>1:
                while line[i]!=-1:
                    if line[i]==startrow[-1]:
                        sepDatabaseSame[startrow[-1]].append(line[i+1:])
                        addLine=1
                        break
                    i+=1
                if addLine!=1:
                    ok=[]
                    while i <len(line):
                        if line[i]==-1:
                            ok=[]
                        elif line[i]==startrow[-1]:
                            ok.append("sk1")
                        for x in give:
                            if x==line[i]:
                                ok.append(x)
                        if len(ok)==1+len(give):
                            sepDatabaseSame[startrow[-1]].append(line[i+1:])
                            break
                        i+=1
        startrow2=[startrow[0]]
        startrow.append(-1)
        if str(startrow) not in self._finalPatterns.keys():
                self.makeNextSame(sepDatabaseSame,startrow2)
        elif self._finalPatterns[str(startrow)]<len(sepDatabaseSame[startrow[-2]]):
            self.makeNextSame(sepDatabaseSame,startrow2)
        return sepDatabaseSame[startrow[-2]]

    def getSameSeq(self,startrow):
        """
         To get words in latest sequence
        :param startrow:
         the patterns get before
        :return:
        """
        give = []
        newrow = startrow.copy()
        while newrow[-1] != -1:
            y = newrow.pop()
            give.append(y)
        return give


    def makeSeqDatabaseSame(self,database,startrow):
        """
            To make sequence dataset list which start from same word(head). It was stored only 1 from 1 line.
            And it separated by having head in latest sequence of startrow or not.

            :param database:
                    To store the transactions of a database in list
            :param startrow: list
                    the patterns get before

            """
        seqDatabase={}
        seqDatabaseSame={}
        seqLength=startrow.count(-1)+1
        for line in database:
            if len(line)>1:
                alreadyInLine=[]
                i = 0
                while line[i] != -1:
                        if line[i] not in seqDatabaseSame:
                            if -1 in startrow:
                                give=self.getSameSeq(startrow)
                            else:
                                give=startrow.copy()
                            newrow= [startrow[-1], line[i]]
                            seqDatabaseSame[line[i]] = self.serchSame(database, newrow,give)

                        i += 1
                same=0
                seqCount=0
                if self._maxLength>seqLength:
                    while len(line)>i and self._maxGap>seqCount:
                        if line[i]!=-1:
                            if line[i] not in alreadyInLine:
                                if line[i] not in seqDatabase:
                                    seqDatabase[line[i]]=[]
                                seqDatabase[line[i]].append(line[i + 1:])
                                alreadyInLine.append(line[i])
                            if line[i]==startrow[-1]:
                                same=1


                            elif same==1 and line[i] not in seqDatabaseSame:
                                if -1 in startrow:
                                    give=self.getSameSeq(startrow)
                                else:
                                    give=startrow.copy()
                                newrow= [startrow[-1], line[i]]
                                seqDatabaseSame[line[i]] = self.serchSame(database, newrow,give)

                        else:
                            same=0
                            seqCount+=1
                        i+=1


        if len(seqDatabase)!=0:
            self.makeNext(seqDatabase,startrow)
        if len(seqDatabaseSame)!=0:
            self.makeNextSame(seqDatabaseSame,startrow)



    def startMine(self):
        """
            Frequent pattern mining process will start from here
        """
        self._Database = []
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._Database=self.makeSupDatabase(self._Database,"")
        self._minSup = self._convert(self._minSup)
        self._mapNeighbours()
        self.makeSeqDatabaseFirst(self._Database)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Frequent patterns were generated successfully using Apriori algorithm ")

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process
        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe
        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def save(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file
        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process
        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        print("Total number of Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) >= 5 and len(_ab._sys.argv) <= 8:
        if len(_ab._sys.argv) == 8:
            _ap = PrefixSpan(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4],_ab._sys.argv[5],_ab._sys.argv[7],_ab._sys.argv[8])
        if len(_ab._sys.argv) == 7:
            _ap = PrefixSpan(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4],_ab._sys.argv[5],_ab._sys.argv[7])
        if len(_ab._sys.argv) == 6:
            _ap = PrefixSpan(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4],_ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = PrefixSpan(_ab._sys.argv[1], _ab._sys.argv[3],_ab._sys.argv[4])
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Frequent Patterns:", len(_Patterns))
        _ap.savePatterns(_ab._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        _ap = PrefixSpan('testdayo.txt',"testN.txt",2, ' ')
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
        print("Total number of Frequent Patterns:", len(_Patterns))
        print("Error! The number of input parameters do not match the total number of parameters provided")
        _ap.save("priOut2.txt")



# TemporalDatabase is a collection of timestamps and along with data at particular time.
#
#  **Importing this algorithm into a python program**
#
#             from PAMI.extras.syntheticDataGenerator import TemporalDatabase as db
#
#             temporalDB = db.TemporalDatabase(numOfTransactions, avgTransactionLength, numItems, outFileName, percentage, sep, occurrenceProbabilityAtSameTimestamp, occurrenceProbabilityToSkipSubsequentTimestamp)
#
#             temporalDB.create()
#
#


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pandas as pd
import numpy as np
import sys


class TemporalDatabase:
    """
    :Description: - creates a temporal database with required parameter (e.g.,databaseSize, avgItemsPerTransaction, numItems and outputFile).
                  - output can be printed in two ways either in text file or dataframe depending on the input type.

    :Attributes:

        :param databaseSize: int
            number of transactions

        :param avgItemsPerTransaction: int
            average length of transactions

    This class generates a temporal database based on the given parameters and provides
    options to output the database in either a text file or a DataFrame format.

    **Importing this algorithm into a Python program**


        from PAMI.extras.syntheticDataGenerator import TemporalDatabase as db

        temporalDB = db.TemporalDatabase(numOfTransactions, avgTransactionLength, numItems, outFileName, percentage, sep, occurrenceProbabilityAtSameTimestamp, occurrenceProbabilityToSkipSubsequentTimestamp)

        temporalDB.create()


    **Methods to execute code on terminal**


    Format:

        (.venv) $ python3 TemporalDatabase.py <numOfTransactions> <avgLenOfTransactions> <numItems> <outputFile> <percentage> <sep> <typeOfFile> <occurrenceProbabilityAtSameTimestamp> <occurrenceProbabilityToSkipSubsequentTimestamp>


    Example Usage:

        (.venv) $ python3 TemporalDatabase.py 50 10 100 temporal.txt 50 \t database 0.1 0.1


    :param numOfTransactions: int
        Number of transactions to generate.

    :param avgLenOfTransactions: int
        Average length of transactions.

    :param numItems: int
        Number of items in the database.

    :param outputFile: str
        Name of the output file for the database.

    :param percentage: int
        Percentage for the coin toss to decide if a transaction will be included in the output.
        If the value is greater than 1, it is treated as a percentage (i.e., 50 for 50%).

    :param sep: str
        Separator for the output file (default is tab).

    :param typeOfFile: str
        Type of output file. Can be 'database' for a text file or 'dataframe' for a DataFrame output.

    :param occurrenceProbabilityAtSameTimestamp: float
        Probability that a new transaction will occur at the same timestamp as the previous one.

    :param occurrenceProbabilityToSkipSubsequentTimestamp: float
        Probability that the timestamp will be skipped for subsequent transactions.
    """

    def __init__(self, numOfTransactions: int, avgLenOfTransactions: int,
                 numItems: int, outputFile: str, percentage: int = 50,
                 sep: str = '\t', typeOfFile: str = "Database",
                 occurrenceProbabilityAtSameTimestamp: float = 0.1,
                 occurrenceProbabilityToSkipSubsequentTimestamp: float = 0.1) -> None:
        """
        Initialize the TemporalDatabase with required parameters.

        :param numOfTransactions: Number of transactions to generate.
        :param avgLenOfTransactions: Average length of transactions.
        :param numItems: Number of items in the database.
        :param outputFile: Name of the output file for the database.
        :param percentage: Percentage for the coin toss to include transactions.
        :param sep: Separator for the output file.
        :param typeOfFile: Type of output file ('database' or 'dataframe').
        :param occurrenceProbabilityAtSameTimestamp: Probability for same timestamp.
        :param occurrenceProbabilityToSkipSubsequentTimestamp: Probability to skip subsequent timestamp.
        """

        self.databaseSize = databaseSize
        self.avgItemsPerTransaction = avgItemsPerTransaction
        self.numItems = numItems
        self.outputFile = outputFile
        if percentage > 1:
            self.percentage = percentage / 100
        else:
            self.percentage = percentage
        self.sep = sep
        self.typeOfFile = typeOfFile.lower()
        self.occurrenceProbabilityAtSameTimestamp = occurrenceProbabilityAtSameTimestamp
        self.occurrenceProbabilityToSkipSubsequentTimestamp = occurrenceProbabilityToSkipSubsequentTimestamp

    def getFileName(self) -> str:
        """
        Returns the name of the output file.

        :return: Output file name.
        """
        return self.outputFile

    def getDatabaseAsDataFrame(self) -> pd.DataFrame:
        """
        Returns the database as a DataFrame.

        :return: pd.DataFrame containing the temporal database.
        """
        return self.df

    def performCoinFlip(self, probability: float) -> bool:
        """
        Perform a coin flip with the given probability.

        :param probability: Probability of the coin landing heads (i.e., the event occurring).
        :return: True if the coin lands heads, False otherwise.
        """
        result = np.random.choice([0, 1], p=[1 - probability, probability])
        return result == 1

    def tuning(self, array, sumRes) -> list:
        """
        Tune the array to ensure that the sum of the values equals sumRes.

        :param array: List of values to be tuned.
        :type array: list
        :param sumRes: Target sum for the array values.
        :type sumRes: int
        :return: Tuned list of values.
        """
        values = np.random.randint(1, self.numItems, len(array))

        while np.sum(values) != sumRes:
            if np.sum(values) > sumRes:
                maxIndex = np.argmax(values)
                values[maxIndex] -= 1
            else:
                minIndex = np.argmin(values)
                values[minIndex] += 1

        for i in range(len(array)):
            array[i][1] = values[i]

        return array

    def create(self) -> None:
        """
        Create the temporal database or DataFrame based on the specified type of file.
        """
        db = []
        lineSize = []

        self.current_timestamp = 0  # Initialize current timestamp

        for i in range(self.numOfTransactions):
            if self.performCoinFlip(self.occurrenceProbabilityAtSameTimestamp):
                timestamp = self.current_timestamp
            else:
                if self.performCoinFlip(self.occurrenceProbabilityToSkipSubsequentTimestamp):
                    self.current_timestamp += 2
                else:
                    self.current_timestamp += 1
                timestamp = self.current_timestamp

            db.append([timestamp])
            if self.performCoinFlip(self.percentage):
                lineSize.append([i, 0])

        sumRes = self.numOfTransactions * self.avgLenOfTransactions

        self.tuning(lineSize, sumRes)

        for i in range(len(lineSize)):
            if lineSize[i][1] > self.numItems:

                raise ValueError(
                    "Error: Either increase numItems or decrease avgLenOfTransactions or modify percentage")
            line = np.random.choice(range(1, self.numItems + 1), lineSize[i][1], replace=False)
            db[lineSize[i][0]].extend(line)

        if self.typeOfFile == "database":
            with open(self.outputFile, "w") as outFile:
                for line in db:
                    outFile.write(self.sep.join(map(str, line)) + '\n')

        if self.typeOfFile == "dataframe":
            data = {
                'timestamp': [line[0] for line in db],
                'transactions': pd.Series([line[1:] for line in db])
            }
            self.df = pd.DataFrame(data)

        print("Temporal database created successfully")


if __name__ == '__main__':
    if len(sys.argv) != 10:
        print("Usage: python TemporalDatabase.py <numOfTransactions> <avgLenOfTransactions> <numItems> <outputFile> <percentage> <sep> <typeOfFile> <occurrenceProbabilityAtSameTimestamp> <occurrenceProbabilityToSkipSubsequentTimestamp>")
        sys.exit(1)

    obj = TemporalDatabase(
        int(sys.argv[1]), int(sys.argv[2]), int(sys.argv[3]), sys.argv[4],
        percentage=int(sys.argv[5]), sep=sys.argv[6], typeOfFile=sys.argv[7],
        occurrenceProbabilityAtSameTimestamp=float(sys.argv[8]),
        occurrenceProbabilityToSkipSubsequentTimestamp=float(sys.argv[9])
    )
    obj.create()



# Top - K is and algorithm to discover top frequent patterns in a transactional database.
#
# **Importing this algorithm into a python program**
#
#             import PAMI.frequentPattern.topK.FAE as alg
#
#             iFile = 'sampleDB.txt'
#
#             K = 2
#
#             obj = alg.FAE(iFile, K)
#
#             obj.mine()
#
#             topKFrequentPatterns = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(topKFrequentPatterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternInDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PAMI.frequentPattern.topk import abstract as _ab
from deprecated import deprecated


class FAE(_ab._frequentPatterns):
    """
    **About this algorithm**

    :**Description**: Top - K is and algorithm to discover top frequent patterns in a transactional database.

    :**Reference**:   Zhi-Hong Deng, Guo-Dong Fang: Mining Top-Rank-K Frequent Patterns: DOI: 10.1109/ICMLC.2007.4370261 · Source: IEEE Xplore https://ieeexplore.ieee.org/document/4370261

    :**Parameters**:    - **iFile** (*str or URL or dataFrame*) -- *Name of the Input file to mine complete set of frequent patterns.*
                        - **oFile** (*str*) -- *Name of the output file to store complete set of frequent patterns.*
                        - **k** (*int*) -- *User specified count of top frequent patterns.*
                        - **minimum** (*int*) -- *Minimum number of frequent patterns to consider in analysis.*
                        - **sep** (*str*) -- *This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.*

    :**Attributes**:    - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the completion time of the mining process.*
                        - **finalPatterns** (*dict*) -- *Storing the complete set of patterns in a dictionary variable.*
                        - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** (*float*) -- *To store the total amount of RSS memory consumed by the program.*

    **Execution methods**

    **Terminal command**

    .. code-block:: console

      Format:

      (.venv) $ python3 FAE.py <inputFile> <outputFile> <K>

      Example Usage:

      (.venv) $ python3 FAE.py sampleDB.txt patterns.txt 10.0

    .. note:: k will be considered as count of top frequent patterns to consider in analysis.

    **Calling from a python program**

    .. code-block:: python

            import PAMI.frequentPattern.topK.FAE as alg

            iFile = 'sampleDB.txt'

            K = 2

            obj = alg.FAE(iFile, K)

            obj.mine()

            topKFrequentPatterns = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(topKFrequentPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)


    **Credits:**

    The complete program was written by P. Likhitha  and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """

    _startTime = float()
    _endTime = float()
    _k = int()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _tidList = {}
    _minimum = int()

    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """

        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._Database = self._iFile['Transactions'].tolist()
                self._Database = [x.split(self._sep) for x in self._Database]
            else:
                print("The column name should be Transactions and each line should be separated by tab space or a seperator specified by the user")

            # print(self.Database)
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def _frequentOneItem(self):
        """
        Generating one frequent patterns
        """
        candidate = {}
        self._tidList = {}
        for i in range(len(self._Database)):
            for j in self._Database[i]:
                if j not in candidate:
                    candidate[j] = 1
                    self._tidList[j] = [i]
                else:
                    candidate[j] += 1
                    self._tidList[j].append(i)
        self._finalPatterns = {}
        plist = [key for key, value in sorted(candidate.items(), key=lambda x: x[1], reverse=True)]
        self._tidList = {k: frozenset(v) for k, v in self._tidList.items()}
        for i in plist:
            if len(self._finalPatterns) >= self._k:
                break
            else:
                self._finalPatterns[i] = candidate[i]
        self._minimum = min([self._finalPatterns[i] for i in self._finalPatterns.keys()])
        plist = list(self._finalPatterns.keys())
        return plist

    def _save(self, prefix, suffix, tidSetI):
        """
        Saves the patterns that satisfy the periodic frequent property.

        :param prefix: the prefix of a pattern
        :type prefix: list
        :param suffix: the suffix of a patterns
        :type suffix: list
        :param tidSetI: the timestamp of a patterns
        :type tidSetI: list
        """

        if prefix is None:
            prefix = suffix
        else:
            prefix = prefix + suffix
        val = len(tidSetI)
        sample = str()
        # for i in prefix:
        #     sample = sample + i + "\t"
        sample = "\t".join(prefix)
        if len(self._finalPatterns) < self._k:
            if val > self._minimum:
                self._finalPatterns[sample] = val
                self._finalPatterns = {k: v for k, v in sorted(self._finalPatterns.items(), key=lambda item: item[1], reverse=True)}
                self._minimum = min([i for i in self._finalPatterns.values()])
        else:
            for x, y in sorted(self._finalPatterns.items(), key=lambda x: x[1]):
                if val > y:
                    del self._finalPatterns[x]
                    self._finalPatterns[sample] = val
                    self._finalPatterns = {k: v for k, v in
                                              sorted(self._finalPatterns.items(), key=lambda item: item[1],
                                                     reverse=True)}
                    self._minimum = min([i for i in self._finalPatterns.values()])
                    return

    def _Generation(self, prefix, itemSets, tidSets):
        """
        Equivalence class is followed  and checks for the patterns generated for periodic-frequent patterns.

        :param prefix:  main equivalence prefix
        :type prefix: periodic-frequent item or pattern
        :param itemSets: patterns which are items combined with prefix and satisfying the periodicity and frequent with their timestamps
        :type itemSets: list
        :param tidSets: timestamps of the items in the argument itemSets
        :type tidSets: list
        """
        if len(itemSets) == 1:
            i = itemSets[0]
            tidI = tidSets[0]
            self._save(prefix, [i], tidI)
            return
        for i in range(len(itemSets)):
            itemI = itemSets[i]
            if itemI is None:
                continue
            tidSetI = tidSets[i]
            classItemSets = []
            classTidSets = []
            itemSetX = [itemI]
            for j in range(i + 1, len(itemSets)):
                itemJ = itemSets[j]
                tidSetJ = tidSets[j]
                y = tidSetI.intersection(tidSetJ)
                if len(y) >= self._minimum:
                    classItemSets.append(itemJ)
                    classTidSets.append(y)
            newPrefix = list(set(itemSetX)) + prefix
            self._Generation(newPrefix, classItemSets, classTidSets)
            self._save(prefix, list(set(itemSetX)), tidSetI)

    def _convert(self, value):
        """
        to convert the type of user specified minSup value

        :param value: user specified minSup value
        :type value: int or float or str
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = ((len(self._Database)) * value)
            else:
                value = int(value)
        return value

    @deprecated("It is recommended to use 'mine()' instead of 'startMine()' for mining process. Starting from January 2025, 'startMine()' will be completely terminated.")
    def startMine(self):
        """
        TopK Frequent pattern mining process will start from here
        """
        self.mine()

    def mine(self):
        """
        TopK Frequent pattern mining process will start from here
        """
        self._startTime = _ab._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._k is None:
            raise Exception("Please enter the Minimum Support")
        self._creatingItemSets()
        self._k = self._convert(self._k)
        plist = self._frequentOneItem()
        for i in range(len(plist)):
            itemI = plist[i]
            tidSetI = self._tidList[itemI]
            itemSetX = [itemI]
            itemSets = []
            tidSets = []
            for j in range(i + 1, len(plist)):
                itemJ = plist[j]
                tidSetJ = self._tidList[itemJ]
                y1 = tidSetI.intersection(tidSetJ)
                if len(y1) >= self._minimum:
                    itemSets.append(itemJ)
                    tidSets.append(y1)
            self._Generation(itemSetX, itemSets, tidSets)
        print(" TopK frequent patterns were successfully generated using FAE algorithm.")
        self._endTime = _ab._time.time()
        self._memoryUSS = float()
        self._memoryRSS = float()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """

        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        # dataframe = {}
        # data = []
        # for a, b in self._finalPatterns.items():
        #     data.append([a.replace('\t', ' '), b])
        #     dataframe = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])

        dataframe = _ab._pd.DataFrame(list([[x.replace('\t', ' '), y] for x,y in self._finalPatterns.items()]), columns=['Patterns', 'Support'])

        return dataframe

    def save(self, outFile):
        """

        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file

        :type outFile: csvfile

        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.strip() + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Top K Frequent  Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = FAE(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = FAE(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.startMine()
        _ap.mine()
        print("Top K Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")




from abc import ABC, abstractmethod
from .graph import Graph
from .DFSCode import DfsCode
from .DFSThread import DfsThread
from .frequentSubgraph import FrequentSubgraph
from .vertex import Vertex
from .edge import Edge
from .extendedEdge import ExtendedEdge
from .sparseTriangularMatrix import SparseTriangularMatrix
from queue import PriorityQueue
import time
import math
import matplotlib.pyplot as plt
import psutil as _psutil
import os as _os
import time


class _TKG(ABC):

    @abstractmethod
    def mine(self):
        """
        Run the tkg algorithm.
        """
        pass

    @abstractmethod
    def readGraphs(self, path):
        """
        Read graphs from a file.
        :param path: Path to the file containing the graphs.
        """
        pass

    @abstractmethod
    def save(self, outputPath):
        """
        Write the result of the gSpan algorithm to a file.
        :param outputPath: Path to the output file.
        """
        pass

    @abstractmethod
    def gSpan(self, graphDb, outputFrequentVertices):
        """
        The main gSpan function to find frequent subgraphs.
        :param graphDb: The database of graphs to mine.
        :param outputFrequentVertices: Boolean indicating whether to output single vertices as subgraphs.
        """
        pass


